#define MODULE
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <asm/io.h>

#include <rtai.h>

#define LOCAL_TIMER_VECTOR  0x41

#define COUNT 4E9

const int SECS = 5;

MODULE_PARM(SECS, "i");

static int RESET_COUNT;

static void calibrate(void)
{
	unsigned long temp;
	static int count = 0, gcount = -1;
	static RTIME tbase;
	long linux_cr0, fpu_reg[27];
	double freq;
	union {unsigned long long time; unsigned long time_lh[2]; } apic;
	apic.time = apic_read(APIC_TMCCT);
	if (gcount < 0) {
		tbase = apic.time;
	}
	gcount++;
	if (++count == RESET_COUNT) {
		apic.time = tbase - apic.time;
		__asm__ __volatile__("movl %%cr0,%%eax": "=a" (linux_cr0): : "ax");
		__asm__ __volatile__("clts; fnsave %0": "=m" (fpu_reg));
		freq =  (double)apic.time_lh[1]*(double)0x100000000LL + (double)apic.time_lh[0];
		count =  (freq*CLOCK_TICK_RATE)/(((double)gcount)*LATCH) + 0.4999999999999;
		__asm__ __volatile__("frstor %0" : "=m" (fpu_reg));
		__asm__ __volatile__("movl %%eax,%%cr0": :"a" (linux_cr0): "ax");
		printk("\n->>> MEASURED APIC_FREQ: %d (Hz) (%d s) <<<-\n", count, gcount/100 + 1);
		count = 0;
	}
	temp = (apic_read(APIC_ICR) & (~0xCDFFF)) |
	       (APIC_DEST_DM_FIXED | APIC_DEST_ALLINC | LOCAL_TIMER_VECTOR);
	apic_write(APIC_ICR, temp);
}

static void just_ret(void)
{
	return;
}

int init_module(void)
{
	RESET_COUNT = SECS*100;
	rt_mount_rtai();
	rt_assign_irq_to_cpu(TIMER_8254_IRQ, hard_cpu_id());
	rt_request_timer(just_ret, COUNT, 1);
	rt_request_global_irq(TIMER_8254_IRQ, calibrate);
	printk("\n->>> HERE WE GO (PRINTING EVERY %d SECONDS) <<<-\n\n", RESET_COUNT/100);
	return 0;
}

void cleanup_module(void)
{
	rt_free_timer();
	rt_reset_irq_to_sym_mode(TIMER_8254_IRQ);
	rt_free_global_irq(TIMER_8254_IRQ);
	rt_umount_rtai();
}
