#define MODULE

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/atomic.h>
#include <linux/malloc.h>

#include <rtai.h>
#include <rtai_sched.h>
#include <rtai_fifos.h>


#define LPT 0x378

#define TIMERTICKS 500000

#define CMDF0 0

//#define ONE_SHOT

static RT_TASK Slow_Task;
static RT_TASK Fast_Task;

static int cpu_used[NR_RT_CPUS];

static void Slow_Thread(int t)
{
        static struct {char task, susres; int flags; RTIME time;} msg = {'S',};                      
        while (1){  
		cpu_used[hard_cpu_id()]++;
                msg.time = rt_get_cpu_time_ns();
                msg.susres = 'r'; 
		hard_save_flags(msg.flags);
                rtf_put(CMDF0, &msg, sizeof(msg)); 

                rt_busy_sleep(11*TIMERTICKS);

                msg.time = rt_get_cpu_time_ns();
                msg.susres = 's'; 
		hard_save_flags(msg.flags);
                rtf_put(CMDF0, &msg, sizeof(msg));

                rt_task_wait_period();                                        
        }
}                                        

static void Fast_Thread(int t) 
{                             
        static struct {char task, susres; int flags; RTIME time;} msg = {'F',};

        while (1) {
		cpu_used[hard_cpu_id()]++;
                msg.time = rt_get_time_ns();
                msg.susres = 'r'; 
		hard_save_flags(msg.flags);
                rtf_put(CMDF0, &msg, sizeof(msg)); 

                rt_busy_sleep(2*TIMERTICKS);

                msg.time = rt_get_time_ns();
                msg.susres = 's'; 
		hard_save_flags(msg.flags);
                rtf_put(CMDF0, &msg, sizeof(msg));

                rt_task_wait_period();                                        
        }                      
}
                                                                                   
int init_module(void)
{                   
        RTIME tick_period;
        RTIME now;                                                                 
	rtf_create_using_bh(CMDF0, 20000, 0);
	rt_task_init(&Fast_Task, Fast_Thread, 0, 2000, 0, 0, 0);
	rt_task_init(&Slow_Task, Slow_Thread, 0, 2000, 1, 0, 0);

#ifdef ONE_SHOT
	rt_set_oneshot_mode();
#endif
	tick_period = 4*start_rt_timer(nano2count(TIMERTICKS));
	now = rt_get_time();
	rt_task_make_periodic(&Fast_Task, now + tick_period,  tick_period);
	rt_task_make_periodic(&Slow_Task, now + tick_period,  6*tick_period);

	return 0;
}

void cleanup_module(void)
{
	int cpuid;
	stop_rt_timer();
	rt_busy_sleep(10000000);
	rtf_destroy(CMDF0);
	rt_task_delete(&Slow_Task);
	rt_task_delete(&Fast_Task);
	printk("\n\nCPU USE SUMMARY\n");
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		printk("# %d -> %d\n", cpuid, cpu_used[cpuid]);
	}
	printk("END OF CPU USE SUMMARY\n\n");
}
