#define TICK 500000 //ns (!!!!! CAREFULL NEVER GREATER THAN 1E7 !!!!!)

#define APIC_TIMER 1

#define TIMER_FREQ (APIC_TIMER ? FREQ_APIC : FREQ_8254)

#define MODULE
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>

#include <asm/system.h>
#include <asm/smp.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>
#include <asm/desc.h>
#include <asm/delay.h>

#include <rtai_fifos.h>
#include <rtai.h>
#include <rtai_sched.h>

static struct wait_queue *waitq;
static struct task_struct *sleeping_process;

static int srq;

static inline int is_signalled(sigset_t signal, sigset_t blocked)
{
	int i =  _NSIG_WORDS - 1;
	do {
		if ( signal.sig[i] & ~blocked.sig[i] ) {
			return 1;
		}
	} while (i--);
	return 0;
}

long long user_srq_handler(unsigned int whatever)
{
	long long time;

	if (whatever == 1) {
		return llimd(rt_times.periodic_tick, 1E6, TIMER_FREQ);
	}
	sleeping_process = current;
	interruptible_sleep_on(&waitq);
	if (is_signalled(current->signal, current->blocked)) {
		return -ERESTARTSYS;
	}
	time = llimd(rt_times.tick_time, 1E6, TIMER_FREQ);
// let's show how to communicate. Copy to and from user shall allow any kind of
// data interchange and services.
	copy_to_user((long long *)whatever, &time, 8);
	return time;
}

void rtai_srq_handler(void)
{
	if (sleeping_process && sleeping_process == waitq->task &&
			sleeping_process->state == TASK_INTERRUPTIBLE) { 
		sleeping_process = 0;
		wake_up_interruptible(&waitq);
	}
}

static void rt_timer_tick(void)
{
	rt_pend_linux_srq(srq);
	rt_times.tick_time = rt_times.intr_time;
	rt_times.intr_time = rt_times.tick_time + rt_times.periodic_tick;
	if (rt_times.tick_time >= rt_times.linux_time) {
		rt_times.linux_time += rt_times.linux_tick;
		if (!APIC_TIMER) {
			rt_pend_linux_irq(TIMER_8254_IRQ);
		}
	} 
}

int init_module(void)
{
	rt_mount_rtai();
	srq = rt_request_srq(0xcacca, rtai_srq_handler, user_srq_handler);
	rt_request_timer(rt_timer_tick, imuldiv(TICK, TIMER_FREQ, 1E9), APIC_TIMER); 
	return 0;
}

void cleanup_module(void)
{
	rt_free_timer();
	rt_free_srq(srq);
	rt_umount_rtai();
	return;
}
