/*
COPYRIGHT (C) 1999  Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
*/


#define RTAI_LXRT_VECTOR  0xFC

static inline unsigned long long rtai_lxrt(int srq, void *arg)
{
	long long retval;
	__asm__ __volatile__ ("int $0xFC"
	: "=A" (retval) : "a" (srq), "d" (arg) : "%eax", "%edx");
	return retval;
}

#ifndef NAMNUM
#define NAMNUM
static inline unsigned long nam2num(const char *name)
{
        int i = 0;
        union { struct { long num; char eos; } num; char name[5]; } retval;
        if (!name[0]) {
           return 0x454e4f4e;
        }
        retval.num.eos = 0;
        while (i < 4 && (retval.name[i] = name[i])) { i++; }
        return retval.num.num;
}

static inline void num2nam(unsigned long num, char *name)
{
        int i = 0;
        union { struct { long num; char eos; } num; char name[5]; } retval;
        retval.num.num = num;
        retval.num.eos = 0;
        while (i < 4 && (name[i] = retval.name[i])) { i++; }
}
#endif

#define YIELD		 	0
#define SUSPEND		 	1
#define RESUME		 	2
#define MAKE_PERIODIC	 	3
#define WAIT_PERIOD	 	4
#define SLEEP		 	5
#define SLEEP_UNTIL	 	6
#define START_TIMER	 	7
#define STOP_TIMER	 	8
#define GET_TIME	 	9
#define COUNT2NANO		10
#define NANO2COUNT		11
#define SEM_SIGNAL		12
#define SEM_WAIT		13
#define SEM_WAIT_IF		14
#define SEM_WAIT_UNTIL		15
#define SEM_WAIT_DELAY		16
#define BUSY_SLEEP		17
#define SENDMSG			18
#define SEND_IF			19
#define SEND_UNTIL		20
#define SEND_TIMED		21
#define RECEIVEMSG		22
#define RECEIVE_IF		23
#define RECEIVE_UNTIL		24
#define RECEIVE_TIMED		25
#define RPCMSG			26
#define RPC_IF			27
#define RPC_UNTIL		28
#define RPC_TIMED		29
#define ISRPC			30
#define RETURNMSG		31
#define SET_PERIODIC_MODE	32
#define SET_ONESHOT_MODE	33
#define SIGNAL_HANDLER 		34
#define TASK_USE_FPU		35
#define LINUX_USE_FPU		36
#define PREEMPT_ALWAYS		37
#define GET_TIME_NS		38
#define GET_CPU_TIME_NS		39

#define GET_ADR	       	100
#define GET_NAME       	101
#define TASK_INIT      	102
#define TASK_DELETE    	103
#define SEM_INIT       	104
#define SEM_DELETE	105

#ifdef MODULE

extern int rt_register(unsigned long name, void *adr);
extern int rt_drg_on_name(unsigned long name);
extern int rt_drg_on_adr(void *adr);
extern unsigned long rt_get_name(void *adr);
extern void *rt_get_adr(unsigned long name);

#ifdef LXRT_MODULE
#define SRQ(x) ((x) >> 24)
#define NARG(x) ((x) & 0xFFFFFF)

static struct { int type; void *fun; } rt_fun[] = { 
	{  1, rt_task_yield },			//   0  
	{  1, rt_task_suspend },		//   1
	{  1, rt_task_resume },			//   2
	{  1, rt_task_make_periodic },		//   3
	{  1, rt_task_wait_period },		//   4
	{  1, rt_sleep },			//   5
	{  1, rt_sleep_until },			//   6
	{  0, start_rt_timer },			//   7
	{  0, stop_rt_timer },			//   8
	{  0, rt_get_time },			//   9
	{  0, count2nano },			//  10
	{  0, nano2count },			//  11
	{  1, rt_sem_signal },			//  12
	{  1, rt_sem_wait },			//  13
	{  1, rt_sem_wait_if },			//  14
	{  1, rt_sem_wait_until },		//  15
	{  1, rt_sem_wait_timed },		//  16
	{  0, rt_busy_sleep },			//  17
	{  1, rt_send },			//  18
	{  1, rt_send_if },			//  19
	{  1, rt_send_until },			//  20
	{  1, rt_send_timed },			//  21
	{  2, rt_receive },   			//  22
	{  2, rt_receive_if }, 			//  23
	{  2, rt_receive_until }, 		//  24
	{  2, rt_receive_timed }, 		//  25
	{  3, rt_rpc },   			//  26
	{  3, rt_rpc_if }, 			//  27
	{  3, rt_rpc_until }, 	      		//  28
	{  3, rt_rpc_timed },   		//  29
	{  0, rt_isrpc }, 	 		//  30
	{  1, rt_return },			//  31
	{  0, rt_set_periodic_mode },		//  32
	{  0, rt_set_oneshot_mode },		//  33
	{  0, rt_task_signal_handler },		//  34
	{  0, rt_task_use_fpu },		//  35
	{  0, rt_linux_use_fpu },		//  36
	{  0, rt_preempt_always },		//  37
	{  0, rt_get_time_ns },			//  38
	{  0, rt_get_cpu_time_ns }		//  39
};
#endif

#else

RT_TASK *this_rt_task;

#define SRQ_NARG(srq) (((srq) << 24) | sizeof((arg))/sizeof(int))

static inline void *rt_get_adr(int name)
{
	int arg = name;
	return (void *)(unsigned long)rtai_lxrt(SRQ_NARG(GET_ADR), &arg);
}

static inline unsigned long rt_get_name(void *adr)
{
	void *arg = adr;
	return rtai_lxrt(SRQ_NARG(GET_NAME), &arg);
}

static inline RT_TASK *rt_task_init(int name, int priority, int stack_size)
{
	struct { int name, priority, stack_size; } arg = { name, priority, stack_size };
	if (!this_rt_task) {
		this_rt_task = (RT_TASK *)(unsigned long)rtai_lxrt(SRQ_NARG(TASK_INIT), &arg);
	}
	return this_rt_task;
}

static inline int rt_task_delete(RT_TASK *task)
{
	void *arg = task;
	return rtai_lxrt(SRQ_NARG(TASK_DELETE), &arg);
}

static inline int rt_task_yield(void)
{
	void *arg = this_rt_task;
	return rtai_lxrt(SRQ_NARG(YIELD), &arg) & 0xFFFFFFFF;
}

static inline int rt_task_suspend(RT_TASK *task)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; } arg = { this_rt_task, task };
	return rtai_lxrt(SRQ_NARG(SUSPEND), &arg) & 0xFFFFFFFF;
}

static inline int rt_task_resume(RT_TASK *task)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; } arg = { this_rt_task, task };
	return rtai_lxrt(SRQ_NARG(RESUME), &arg) & 0xFFFFFFFF;
}

static inline int rt_task_make_periodic(RT_TASK *task, RTIME start_time, RTIME period)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; RTIME start_time, period; } arg = { this_rt_task, task, start_time, period };
	return rtai_lxrt(SRQ_NARG(MAKE_PERIODIC), &arg) & 0xFFFFFFFF;
}

static inline void rt_task_wait_period(void)
{
	void *arg = this_rt_task;
	rtai_lxrt(SRQ_NARG(WAIT_PERIOD), &arg);
}

static inline void rt_sleep(RTIME delay)
{
	struct { RT_TASK *this_rt_task; RTIME delay; } arg = { this_rt_task, delay };
	rtai_lxrt(SRQ_NARG(SLEEP), &arg);
}

static inline void rt_sleep_until(RTIME time)
{
	struct { RT_TASK *this_rt_task; RTIME time; } arg = { this_rt_task, time };
	rtai_lxrt(SRQ_NARG(SLEEP_UNTIL), &arg);
}

static inline int start_rt_timer(int period)
{
	int arg = period;
	return (int)rtai_lxrt(SRQ_NARG(START_TIMER), &arg);
}

static inline void stop_rt_timer(void)
{
	char arg = 0;
	rtai_lxrt(SRQ_NARG(STOP_TIMER), &arg);
}

static inline RTIME rt_get_time(void)
{
	char arg = 0;
	return rtai_lxrt(SRQ_NARG(GET_TIME), &arg);
}

static inline RTIME rt_count2nano(RTIME count)
{
	RTIME arg = count;
	return rtai_lxrt(SRQ_NARG(COUNT2NANO), &arg);
}

static inline RTIME rt_nano2count(RTIME nano)
{
	RTIME arg = nano;
	return rtai_lxrt(SRQ_NARG(NANO2COUNT), &arg);
}

static inline SEM *rt_sem_init(int name, int value)
{
	struct { int name, value; } arg = { name, value };
	return (SEM *)(unsigned long)rtai_lxrt(SRQ_NARG(SEM_INIT), &arg);
}

static inline int rt_sem_delete(SEM *sem)
{
	void *arg = sem;
	return rtai_lxrt(SRQ_NARG(SEM_DELETE), &arg);
}

static inline int rt_sem_signal(SEM *sem)
{
	struct { RT_TASK *this_rt_task; SEM *sem; } arg = { this_rt_task, sem };
	return rtai_lxrt(SRQ_NARG(SEM_SIGNAL), &arg) & 0xFFFFFFFF;
}

static inline int rt_sem_wait(SEM *sem)
{
	struct { RT_TASK *this_rt_task; SEM *sem; } arg = { this_rt_task, sem };
	return rtai_lxrt(SRQ_NARG(SEM_WAIT), &arg) & 0xFFFFFFFF;
}

static inline int rt_sem_wait_if(SEM *sem)
{
	struct { RT_TASK *this_rt_task; SEM *sem; } arg = { this_rt_task, sem };
	return rtai_lxrt(SRQ_NARG(SEM_WAIT_IF), &arg) & 0xFFFFFFFF;
}

static inline int rt_sem_wait_until(SEM *sem, RTIME time)
{
	struct { RT_TASK *this_rt_task; SEM *sem; RTIME time; } arg = { this_rt_task, sem, time };
	return rtai_lxrt(SRQ_NARG(SEM_WAIT_UNTIL), &arg) & 0xFFFFFFFF;
}

static inline int rt_sem_wait_timed(SEM *sem, RTIME delay)
{
	struct { RT_TASK *this_rt_task; SEM *sem; RTIME delay; } arg = { this_rt_task, sem, delay };
	return rtai_lxrt(SRQ_NARG(SEM_WAIT_DELAY), &arg) & 0xFFFFFFFF;
}

static inline void rt_busy_sleep(int ns)
{
	int arg = ns;
	rtai_lxrt(SRQ_NARG(BUSY_SLEEP), &arg);
}

static inline RT_TASK *rt_send(RT_TASK *task, unsigned int msg)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; unsigned int msg; } arg = { this_rt_task, task, msg };
	return (RT_TASK *)(unsigned long)(rtai_lxrt(SRQ_NARG(SENDMSG), &arg) & 0xFFFFFFFF);
}

static inline RT_TASK *rt_send_if(RT_TASK *task, unsigned int msg)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; unsigned int msg; } arg = { this_rt_task, task, msg };
	return (RT_TASK *)(unsigned long)(rtai_lxrt(SRQ_NARG(SEND_IF), &arg) & 0xFFFFFFFF);
}

static inline RT_TASK *rt_send_until(RT_TASK *task, unsigned int msg, RTIME time)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; unsigned int msg; RTIME time; } arg = { this_rt_task, task, msg, time };
	return (RT_TASK *)(unsigned long)(rtai_lxrt(SRQ_NARG(SEND_UNTIL), &arg) & 0xFFFFFFFF);
}

static inline RT_TASK *rt_send_timed(RT_TASK *task, unsigned int msg, RTIME delay)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; unsigned int msg; RTIME delay; } arg = { this_rt_task, task, msg, delay };
	return (RT_TASK *)(unsigned long)(rtai_lxrt(SRQ_NARG(SEND_TIMED), &arg) & 0xFFFFFFFF);
}

static inline RT_TASK *rt_receive(RT_TASK *task, unsigned int *msg)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; unsigned int *msg; } arg = { this_rt_task, task, msg };
	return (RT_TASK *)(unsigned long)(rtai_lxrt(SRQ_NARG(RECEIVEMSG), &arg) & 0xFFFFFFFF);
}

static inline RT_TASK *rt_receive_if(RT_TASK *task, unsigned int *msg)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; unsigned int *msg; } arg = { this_rt_task, task, msg };
	return (RT_TASK *)(unsigned long)(rtai_lxrt(SRQ_NARG(RECEIVE_IF), &arg) & 0xFFFFFFFF);
}

static inline RT_TASK *rt_receive_until(RT_TASK *task, unsigned int *msg, RTIME time)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; unsigned int *msg; RTIME time; } arg = { this_rt_task, task, msg, time };
	return (RT_TASK *)(unsigned long)(rtai_lxrt(SRQ_NARG(RECEIVE_UNTIL), &arg) & 0xFFFFFFFF);
}

static inline RT_TASK *rt_receive_timed(RT_TASK *task, unsigned int *msg, RTIME delay)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; unsigned int *msg; RTIME delay; } arg = { this_rt_task, task, msg, delay };
	return (RT_TASK *)(unsigned long)(rtai_lxrt(SRQ_NARG(RECEIVE_TIMED), &arg) & 0xFFFFFFFF);
}

static inline RT_TASK *rt_rpc(RT_TASK *task, unsigned int to_do, unsigned int *result)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; unsigned int to_do; unsigned int *result; } arg = { this_rt_task, task, to_do, result };
	return (RT_TASK *)(unsigned long)(rtai_lxrt(SRQ_NARG(RPCMSG), &arg) & 0xFFFFFFFF);
}


static inline RT_TASK *rt_rpc_if(RT_TASK *task, unsigned int to_do, unsigned int *result)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; unsigned int to_do; unsigned int *result; } arg = { this_rt_task, task, to_do, result };
	return (RT_TASK *)(unsigned long)(rtai_lxrt(SRQ_NARG(RPC_IF), &arg) & 0xFFFFFFFF);
}

static inline RT_TASK *rt_rpc_until(RT_TASK *task, unsigned int to_do, unsigned int *result, RTIME time)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; unsigned int to_do; unsigned int *result; RTIME time; } arg = { this_rt_task, task, to_do, result, time };
	return (RT_TASK *)(unsigned long)(rtai_lxrt(SRQ_NARG(RPC_UNTIL), &arg) & 0xFFFFFFFF);
}

static inline RT_TASK *rt_rpc_timed(RT_TASK *task, unsigned int to_do, unsigned int *result, RTIME delay)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; unsigned int to_do; unsigned int *result; RTIME delay; } arg = { this_rt_task, task, to_do, result, delay };
	return (RT_TASK *)(unsigned long)(rtai_lxrt(SRQ_NARG(RPC_TIMED), &arg) & 0xFFFFFFFF);
}

static inline int rt_isrpc(RT_TASK *task)
{
	void *arg = task;
	return rtai_lxrt(SRQ_NARG(ISRPC), &arg);
}

static inline RT_TASK *rt_return(RT_TASK *task, unsigned int result)
{
	struct { RT_TASK *this_rt_task; RT_TASK *task; unsigned int result; } arg = { this_rt_task, task, result };
	return (RT_TASK *)(unsigned long)(rtai_lxrt(SRQ_NARG(RETURNMSG), &arg) & 0xFFFFFFFF);
}

static inline void rt_set_periodic_mode(void)
{
	char arg = 0;
	rtai_lxrt(SRQ_NARG(SET_PERIODIC_MODE), &arg);
}

static inline void rt_set_oneshot_mode(void)
{
	char arg = 0;
	rtai_lxrt(SRQ_NARG(SET_ONESHOT_MODE), &arg);
}

static inline int rt_task_signal_handler(RT_TASK *task, void (*handler)(void))
{
	struct { RT_TASK *task; void (*handler)(void); } arg = { task, handler };
	return rtai_lxrt(SRQ_NARG(SIGNAL_HANDLER), &arg);
}

static inline int rt_task_use_fpu(RT_TASK *task, int use_fpu_flag)
{
	struct { RT_TASK *task; int use_fpu_flag; } arg = { task, use_fpu_flag };
	return rtai_lxrt(SRQ_NARG(TASK_USE_FPU), &arg);
}

static inline int rt_linux_use_fpu(int use_fpu_flag)
{
	struct { int use_fpu_flag; } arg = { use_fpu_flag };
	return rtai_lxrt(SRQ_NARG(LINUX_USE_FPU), &arg);
}

static inline void rt_preempt_always(int yes_no)
{
	struct { int yes_no; } arg = { yes_no };
	rtai_lxrt(SRQ_NARG(PREEMPT_ALWAYS), &arg);
}

static inline RTIME rt_get_time_ns(void)
{
	char arg = 0;
	return rtai_lxrt(SRQ_NARG(GET_TIME_NS), &arg);
}

static inline RTIME rt_get_cpu_time_ns(void)
{
	char arg = 0;
	return rtai_lxrt(SRQ_NARG(GET_CPU_TIME_NS), &arg);
}
#endif
