/*
COPYRIGHT (C) 1999  Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
*/


#define MAX_SLOTS 256  // Set it as you need, but KEEP IT POWER OF 2, ALWAYS!

#define LXRT_MODULE
#include <linux/version.h>
#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/wrapper.h>

#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/io.h>

#include <rtai.h>
#include <rtai_sched.h>
#include <rtai_lxrt.h>

//#define RT_TASK_RESUME
#ifdef  RT_TASK_RESUME
#define LXRT_RESUME(x) { (x)->state = (READY | SUSPENDED); rt_task_resume((x)); }
#else
#define LXRT_RESUME(x) { extern void emuser_trxl(RT_TASK *); emuser_trxl((x)); }
#endif

//#define CONSERVATIVE_LOCK
#ifdef  CONSERVATIVE_LOCK
#define DEFINE_FLAGS    unsigned long flags
#define LXRT_LOCK(x)    flags = rt_spin_lock_irqsave(&(x))
#define LXRT_UNLOCK(x)  rt_spin_unlock_irqrestore(flags, &(x))
#else
#define DEFINE_FLAGS
#define LXRT_LOCK(x)    rt_spin_lock_irq(&(x))
#define LXRT_UNLOCK(x)  rt_spin_unlock_irq(&(x))
#endif

#define __STR(x) #x
#define STR(x) __STR(x)

static void rtai_lxrt_handler(void)
{
	__asm__ __volatile__ (" \
	cld; pushl %es; pushl %ds; pushl %ebp;\n\t \
	pushl %edi; pushl %esi; pushl %ecx;\n\t \
	pushl %ebx; pushl %edx; pushl %eax;\n\t \
	movl $" STR(__KERNEL_DS) ",%ebx; movl %bx,%ds; movl %bx,%es");

	__asm__ __volatile__ ("call "SYMBOL_NAME_STR(lxrt_handler));

	__asm__ __volatile__ (" \
	addl $8,%esp; popl %ebx; popl %ecx; popl %esi;\n\t \
	popl %edi; popl %ebp; popl %ds; popl %es; iret");
}


static struct { unsigned long name; void *adr; } lxrt_list[MAX_SLOTS + 1] = { {0, 0}, };

static spinlock_t list_lock = SPIN_LOCK_UNLOCKED;

static inline int registr(unsigned long name, void *adr)
{
	unsigned long flags;
	int slot;
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (!lxrt_list[slot].name) {
			lxrt_list[slot].name = name;
			lxrt_list[slot].adr  = adr;
			rt_spin_unlock_irqrestore(flags, &list_lock);
			return slot;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

static inline int drg_on_name(unsigned long name)
{
	unsigned long flags;
	int slot;
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (lxrt_list[slot].name == name) {
			lxrt_list[slot] = lxrt_list[0];
			rt_spin_unlock_irqrestore(flags, &list_lock);
			return slot;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

static inline int drg_on_adr(void *adr)
{
	unsigned long flags;
	int slot;
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (lxrt_list[slot].adr == adr) {
			lxrt_list[slot] = lxrt_list[0];
			rt_spin_unlock_irqrestore(flags, &list_lock);
			return slot;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

static inline unsigned long get_name(void *adr)
{
	unsigned long flags;
	int slot;
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (lxrt_list[slot].adr == adr) {
		rt_spin_unlock_irqrestore(flags, &list_lock);
		return lxrt_list[slot].name;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

static inline void *get_adr(unsigned long name)
{
	unsigned long flags;
	int slot;
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (lxrt_list[slot].name == name) {
			rt_spin_unlock_irqrestore(flags, &list_lock);
			return lxrt_list[slot].adr;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

int rt_register(unsigned long name, void *adr)
{
	return registr(name, adr);
}

int rt_drg_on_name(unsigned long name)
{
	return drg_on_name(name);
} 

int rt_drg_on_adr(void *adr)
{
	return drg_on_adr(adr);
} 

unsigned long rt_get_name(void *adr)
{
	return get_name(adr);
} 

void *rt_get_adr(unsigned long name)
{
	return get_adr(name);
}

static struct {int srq, in, out; struct wait_queue **waitq[MAX_SLOTS]; spinlock_t lock;} sysrq;

/* BE CAREFULL TO KEEP THE 3 MACROS BELOW UP TO DATE WITH THE SCHEDULERS */
#define LINUX  0xFFFFF000
#define STK    256      // Keep an eye on this.
#define MAX_RES_ARGS 8  // Keep an eye on this also.

typedef union { RT_TASK *lx[MAX_RES_ARGS]; } ARGS;

static void lxrt_suspend(ARGS args)
{
#define task (args.lx[MAX_RES_ARGS - 1])
	DEFINE_FLAGS;

	__asm__ __volatile__ ("" : "=A" (task->retval) );
	LXRT_LOCK(task->lock);
	task->state = (READY | SUSPENDED | LINUX);
	if (task->lx_proc) { 
		task->lx_proc = 0;
		LXRT_UNLOCK(task->lock);
		LXRT_LOCK(sysrq.lock);
		sysrq.waitq[sysrq.in] = &task->waitq;
		sysrq.in = (sysrq.in + 1) & (MAX_SLOTS - 1);
		rt_pend_linux_srq(sysrq.srq);
		LXRT_UNLOCK(sysrq.lock);
	} else {
		LXRT_UNLOCK(task->lock);
	}
	rt_task_suspend(task);
	printk("!!! GREAT PAIN, THIS POINT CANNOT BE REACHED !!!\n"); 
#undef task
}

static void lxrt_global_sti(void)
{
	rt_release_global_lock();
	hard_sti();
}

static inline long long lxrt_resume(void *fun, int narg, int *arg, int rcv, RT_TASK *task)
{
	struct wait_queue waitq;
	int msg, *msgadr;
	DEFINE_FLAGS;

	copy_from_user(task->stack_top, arg, (narg - 1)*sizeof(int));
	if (rcv) {
		msgadr = (int *)(*(task->stack_top + rcv));
		*(task->stack_top + rcv) = (int)(&msg);
	} else {
		msgadr = 0;
	}
	*(task->stack_top - 1) = (int)lxrt_suspend;
	*(task->stack_top - 2) = (int)fun;
	*(task->stack = task->stack_top - 3) = (int)lxrt_global_sti;
	LXRT_RESUME(task);
	LXRT_LOCK(task->lock);
	if (task->state != (READY | SUSPENDED | LINUX)) {
		task->lx_proc = waitq.task = current;
		current->state = TASK_INTERRUPTIBLE;
		add_wait_queue(&(task->waitq), &waitq);
		LXRT_UNLOCK(task->lock);
		schedule();
		LXRT_LOCK(task->lock);
		remove_wait_queue(&(task->waitq), &waitq);
	}
	LXRT_UNLOCK(task->lock);
	if (msgadr) {
		put_user(msg, msgadr);
	}
	return task->retval;
}

long long lxrt_handler(unsigned int lxsrq, void *arg)
{
	union {unsigned long name; RT_TASK *task; SEM *sem; } arg0;
	unsigned long name;
	int srq;

	get_user(arg0.name, (unsigned long *)arg);
	if ((srq = SRQ(lxsrq)) < GET_ADR) {
		if (rt_fun[srq].type) {
			return lxrt_resume(rt_fun[srq].fun, NARG(lxsrq), &((int *)arg)[1], rt_fun[srq].type - 1, arg0.task);
		}

		switch (NARG(lxsrq)) {
			case 0:
				return ((long long (*)(void))rt_fun[srq].fun)();
			case 1:
				return ((long long (*)(int))rt_fun[srq].fun)(arg0.name);
			case 2: {
				int arg1;
				get_user(arg1, &((int *)arg)[1]);
				return ((long long (*)(int, int))rt_fun[srq].fun)(arg0.name, arg1);
			}
		}
	}

	switch (srq) {

		case GET_ADR: {
			return (unsigned long)get_adr(arg0.name);
		}

		case GET_NAME: {
			return get_name((void *)arg0.name);
		}

		case TASK_INIT: {
			if (get_adr(name = arg0.name)) {
				return 0;
			}
			if ((arg0.task = kmalloc(sizeof(RT_TASK), GFP_KERNEL))) {
			int prio, stk;
				get_user(prio, &((int *)arg)[1]);
				if (prio > RT_LOWEST_PRIORITY) {
					prio = RT_LOWEST_PRIORITY;
				}
				get_user(stk, &((int *)arg)[2]);
				if (!stk) {
					stk = STK;
				}
				if (!rt_task_init(arg0.task, 0, 0, stk, prio, 0, 0)) {
					arg0.task->state = (READY | SUSPENDED | LINUX);
					*arg0.task->stack = (int)arg0.task;
					arg0.task->stack_top = arg0.task->stack - MAX_RES_ARGS;
					arg0.task->lx_proc   = 0;
					arg0.task->waitq     = 0;
					arg0.task->lock.lock = 0;
					if (registr(name, arg0.task)) {
						return arg0.name;
					} else {
						kfree(arg0.task);
					}
				}
			}
			return 0;
		}

		case TASK_DELETE: {
			if (rt_task_delete(arg0.task)) {
				return -EFAULT;
			}
			kfree(arg0.task);
			return drg_on_adr(arg0.task);
		}

		case SEM_INIT: {
			if (get_adr(name = arg0.name)) {
				return 0;
			}
			if ((arg0.sem = kmalloc(sizeof(SEM), GFP_KERNEL))) {
				int ini_cnt;
				get_user(ini_cnt, &((int *)arg)[1]);
				rt_sem_init(arg0.sem, ini_cnt);
				if (registr(name, arg0.sem)) {
					return arg0.name;
				} else {
					kfree(arg0.sem);
				}
			}
			return 0;
		}

		case SEM_DELETE: {
			if (rt_sem_delete(arg0.sem)) {
				return -EFAULT;
			}
			kfree(arg0.sem);
			return drg_on_adr(arg0.sem);
		}
	}
	return 0;
}

static void lxrt_srq_handler(void)
{
	struct wait_queue **waitq;
	DEFINE_FLAGS;

	LXRT_LOCK(sysrq.lock);
	while (sysrq.out != sysrq.in) {
		waitq = sysrq.waitq[sysrq.out];
		sysrq.out = (sysrq.out + 1) & (MAX_SLOTS - 1);
		LXRT_UNLOCK(sysrq.lock);
		wake_up_interruptible(waitq);
		LXRT_LOCK(sysrq.lock);
	
	}
	LXRT_UNLOCK(sysrq.lock);
}

static struct desc_struct sidt;

int init_module(void)
{
	sidt = rt_set_full_intr_vect(RTAI_LXRT_VECTOR, 15, 3, rtai_lxrt_handler);
	if ((sysrq.srq = rt_request_srq(nam2num("LXRT"), lxrt_srq_handler, 0)) < 0) {
		printk("LXRT: no sysrq available.\n");
		return sysrq.srq;
	}
	sysrq.in = sysrq.out = sysrq.lock.lock = 0;
	return 0 ;
}

void cleanup_module(void)
{
	if (rt_free_srq(sysrq.srq) < 0) {
		printk("LXRT: sysrq %d illegal or already free.\n", sysrq.srq);
	}
	rt_reset_full_intr_vect(RTAI_LXRT_VECTOR, sidt);
	return;
}
