//////////////////////////////////////////////////////////////////////////////
//
//      Copyright  1999 Zentropic Computing, All rights reserved
//
// Authors:             Steve Papacharalambous
// Original date:       Thu 15 Jul 1999
//
// pthreads interface for Real Time Linux.
//
//
// The contents of this file are subject to the Mozilla Public
// License Version 1.0 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
//
///////////////////////////////////////////////////////////////////////////////

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>

#include <rtai_pthread.h>



#define RUNNABLE_ON_CPUS 3
#define RUN_ON_CPUS (smp_num_cpus > 1 ? RUNNABLE_ON_CPUS : 1)

#define TIMER_TO_CPU 3             // < 0 || > 1 to maintain a symmetrically processed timer

#define RT_STACK 2000

#define DEFAULT_TICK_PERIOD 20000



void *thread_func(void *arg) {

  int i;
  int r_c;
  int my_policy;
  pthread_t self;
  struct sched_param my_param;


  printk("Starting pthread: %s, id: %ld\n", (char *)arg, pthread_self());

  self = pthread_self();

  if(( r_c = pthread_getschedparam(self, &my_policy, &my_param)) != 0) {
    printk("pthread: %s, pthread_getschedparam error: %d\n", (char *)arg, r_c);
  }

  printk("pthread: %s, running at %s/%d\n", (char *)arg,
          (my_policy == SCHED_FIFO ? "FIFO"
          : (my_policy == SCHED_RR ? "RR"
          : (my_policy == SCHED_OTHER ? "OTHER"
          : "unknown"))),
          my_param.sched_priority);


  printk("pthread: %s, Increasing priority...\n", (char *)arg);

  my_param.sched_priority = 20;
  if(( r_c = pthread_setschedparam(self, SCHED_FIFO, &my_param)) != 0) {
    printk("pthread: %s, pthread_setschedparam error: %d\n", (char *)arg, r_c);
  }

  if(( r_c = pthread_getschedparam(self, &my_policy, &my_param)) != 0) {
    printk("pthread: %s, pthread_getschedparam error: %d\n", (char *)arg, r_c);
  }

  printk("pthread: %s, running at %s/%d\n", (char *)arg,
          (my_policy == SCHED_FIFO ? "FIFO"
          : (my_policy == SCHED_RR ? "RR"
          : (my_policy == SCHED_OTHER ? "OTHER"
          : "unknown"))),
          my_param.sched_priority);


  for(i = 0; i < 5; i++) {

    printk("pthread %s, loop count %d\n", (char *)arg, i);

  }

  pthread_exit("a");

  return(NULL);

}



int init_module(void) {

  int r_c;
  pthread_t thread_id = 0;
  RTIME tick_period;

  printk("pthread test program.\n");

  if(( r_c = pthread_create(&thread_id, NULL, thread_func, "a")) != 0 ) {

    printk("ptest: Thread creation failed - %d\n", r_c);
    return(-1);

  }

  printk("ptest: Created pthread, id = %ld\n", thread_id);

  return(0);

}


void cleanup_module(void) {

  printk("ptest module removed.\n");

}
