//////////////////////////////////////////////////////////////////////////////
//
//      Copyright  1999 Zentropic Computing, All rights reserved
//
// Authors:             Steve Papacharalambous
// Original date:       Thu 15 Jul 1999
//
// pthreads interface for Real Time Linux.
//
//
// The contents of this file are subject to the Mozilla Public
// License Version 1.0 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
//
///////////////////////////////////////////////////////////////////////////////

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>

#include <rtai_pthread.h>



#define RUNNABLE_ON_CPUS 3
#define RUN_ON_CPUS (smp_num_cpus > 1 ? RUNNABLE_ON_CPUS : 1)

#define TIMER_TO_CPU 3             // < 0 || > 1 to maintain a symmetrically processed timer

#define STACK_SIZE 2000

#define TICK_PERIOD 20000


RT_TASK parent_task, child_task;


void child_func(int arg) {

  int i;

  printk("Starting child pthread %d\n", arg);
  for(i = 0; i < 3; i++) {

    printk("child pthread %d, loop count %d\n", arg, i);

  }

  rt_task_delete(rt_whoami());

}



void parent_func(int arg) {

  int i;

  printk("Starting parent pthread %d\n", arg);
  for(i = 0; i < 5; i++) {

    printk("parent pthread %d, loop count %d\n", arg, i);

  }

  rt_task_init(&child_task, child_func, 2, STACK_SIZE, RT_LOWEST_PRIORITY - 1, 1, 0);
  rt_set_runnable_on_cpus(&child_task, RUN_ON_CPUS);
  rt_task_resume(&child_task);
  rt_task_delete(rt_whoami());

}



int init_module(void) {

  RTIME now, tick_period;

  printk("rtaitest test program.\n");

  rt_task_init(&parent_task, parent_func, 1, STACK_SIZE, RT_LOWEST_PRIORITY - 1, 1, 0);
  rt_set_runnable_on_cpus(&parent_task, RUN_ON_CPUS);
  tick_period = start_rt_timer((int)nano2count(TICK_PERIOD));
  rt_assign_irq_to_cpu(TIMER_8254_IRQ, TIMER_TO_CPU);
  rt_linux_use_fpu(1);
  now = rt_get_time();
  rt_task_resume(&parent_task);
  return 0;

}


void cleanup_module(void) {

  rt_reset_irq_to_sym_mode(TIMER_8254_IRQ);
  stop_rt_timer();
  printk("rtaitest module removed.\n");

}
